@@ -16,6 +16,7 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include <stdbool.h>
 #include "command.h"
 #include "init.h"
 #include "quota.h"
@@ -48,20 +49,20 @@ free_help(void)
 static int
 mount_free_space_data(
 	struct fs_path		*mount,
-	__uint64_t		*bcount,
-	__uint64_t		*bused,
-	__uint64_t		*bfree,
-	__uint64_t		*icount,
-	__uint64_t		*iused,
-	__uint64_t		*ifree,
-	__uint64_t		*rcount,
-	__uint64_t		*rused,
-	__uint64_t		*rfree)
+	uint64_t		*bcount,
+	uint64_t		*bused,
+	uint64_t		*bfree,
+	uint64_t		*icount,
+	uint64_t		*iused,
+	uint64_t		*ifree,
+	uint64_t		*rcount,
+	uint64_t		*rused,
+	uint64_t		*rfree)
 {
 	struct xfs_fsop_counts	fscounts;
 	struct xfs_fsop_geom	fsgeo;
 	struct statfs		st;
-	__uint64_t		logsize, count, free;
+	uint64_t		logsize, count, free;
 	int			fd;
 
 	if ((fd = open(mount->fs_dir, O_RDONLY)) < 0) {
@@ -76,34 +77,50 @@ mount_free_space_data(
 		close(fd);
 		return 0;
 	}
-	if ((xfsctl(mount->fs_dir, fd, XFS_IOC_FSGEOMETRY_V1, &fsgeo)) < 0) {
-		perror("XFS_IOC_FSGEOMETRY_V1");
-		close(fd);
-		return 0;
-	}
-	if ((xfsctl(mount->fs_dir, fd, XFS_IOC_FSCOUNTS, &fscounts)) < 0) {
-		perror("XFS_IOC_FSCOUNTS");
-		close(fd);
-		return 0;
+
+	if (!(mount->fs_flags & FS_FOREIGN)) {
+		if ((xfsctl(mount->fs_dir, fd, XFS_IOC_FSGEOMETRY_V1,
+							&fsgeo)) < 0) {
+			perror("XFS_IOC_FSGEOMETRY_V1");
+			close(fd);
+			return 0;
+		}
+		if ((xfsctl(mount->fs_dir, fd, XFS_IOC_FSCOUNTS,
+							&fscounts)) < 0) {
+			perror("XFS_IOC_FSCOUNTS");
+			close(fd);
+			return 0;
+		}
+
+		logsize = fsgeo.logstart ? fsgeo.logblocks : 0;
+		count = (fsgeo.datablocks - logsize) * fsgeo.blocksize;
+		free  = fscounts.freedata * fsgeo.blocksize;
+		*bcount = BTOBB(count);
+		*bfree  = BTOBB(free);
+		*bused  = BTOBB(count - free);
+
+		count = fsgeo.rtextents * fsgeo.rtextsize * fsgeo.blocksize;
+		free  = fscounts.freertx * fsgeo.rtextsize * fsgeo.blocksize;
+		*rcount = BTOBB(count);
+		*rfree  = BTOBB(free);
+		*rused  = BTOBB(count - free);
+	} else {
+		count = st.f_blocks * st.f_bsize;
+		free = st.f_bfree * st.f_bsize;
+		*bcount = BTOBB(count);
+		*bfree  = BTOBB(free);
+		*bused  = BTOBB(count - free);
+
+		*rcount = BTOBB(0);
+		*rfree  = BTOBB(0);
+		*rused  = BTOBB(0);
 	}
 
-	logsize = fsgeo.logstart ? fsgeo.logblocks : 0;
-	count = (fsgeo.datablocks - logsize) * fsgeo.blocksize;
-	free  = fscounts.freedata * fsgeo.blocksize;
-	*bcount = BTOBB(count);
-	*bfree  = BTOBB(free);
-	*bused  = BTOBB(count - free);
 
 	*icount = st.f_files;
 	*ifree  = st.f_ffree;
 	*iused  = st.f_files - st.f_ffree;
 
-	count = fsgeo.rtextents * fsgeo.rtextsize * fsgeo.blocksize;
-	free  = fscounts.freertx * fsgeo.rtextsize * fsgeo.blocksize;
-	*rcount = BTOBB(count);
-	*rfree  = BTOBB(free);
-	*rused  = BTOBB(count - free);
-
 	close(fd);
 	return 1;
 }
@@ -115,15 +132,15 @@ mount_free_space_data(
 static int
 projects_free_space_data(
 	struct fs_path		*path,
-	__uint64_t		*bcount,
-	__uint64_t		*bused,
-	__uint64_t		*bfree,
-	__uint64_t		*icount,
-	__uint64_t		*iused,
-	__uint64_t		*ifree,
-	__uint64_t		*rcount,
-	__uint64_t		*rused,
-	__uint64_t		*rfree)
+	uint64_t		*bcount,
+	uint64_t		*bused,
+	uint64_t		*bfree,
+	uint64_t		*icount,
+	uint64_t		*iused,
+	uint64_t		*ifree,
+	uint64_t		*rcount,
+	uint64_t		*rused,
+	uint64_t		*rfree)
 {
 	fs_quota_stat_t		qfs;
 	fs_disk_quota_t		d;
@@ -209,9 +226,9 @@ free_space(
 	fs_path_t	*path,
 	uint		flags)
 {
-	__uint64_t	bcount, bused, bfree;
-	__uint64_t	icount, iused, ifree;
-	__uint64_t	rcount, rused, rfree;
+	uint64_t	bcount, bused, bfree;
+	uint64_t	icount, iused, ifree;
+	uint64_t	rcount, rused, rfree;
 	char		a[8], s[8], u[8], p[8];
 	int		count;
 
@@ -368,9 +385,10 @@ free_init(void)
 	free_cmd.cfunc = free_f;
 	free_cmd.argmin = 0;
 	free_cmd.argmax = -1;
-	free_cmd.args = _("[-bir] [-hn] [-f file]");
+	free_cmd.args = _("[-bir] [-hN] [-f file]");
 	free_cmd.oneline = _("show free and used counts for blocks and inodes");
 	free_cmd.help = free_help;
+	free_cmd.flags = CMD_FLAG_FOREIGN_OK;
 
 	add_command(&free_cmd);
 }
